package info.textgrid.middleware;

import static org.junit.Assert.assertTrue;
import java.io.IOException;
import java.io.InputStream;
import javax.ws.rs.core.Response;
import org.apache.cxf.helpers.IOUtils;
import org.apache.cxf.jaxrs.client.Client;
import org.apache.cxf.jaxrs.client.JAXRSClientFactory;
import org.apache.cxf.jaxrs.client.WebClient;
import org.apache.cxf.transport.http.HTTPConduit;
import org.apache.cxf.transports.http.configuration.HTTPClientPolicy;
import org.apache.http.HttpStatus;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

/**
 * <p>
 * Some online tests for the DARIAH-DE Repository OAIMPH service.
 * </p>
 * 
 * @author Stefan E. Funk, SUB Göttingen
 */
@Ignore
public class OaiPmhDariahdeOnlineTests {

  // TODO Configure config files for all the different OAI-PMH service instances!

  // **
  // The OAIPMH host to be tested.
  // **

  // The OAIPMH host to be tested.
  // private static String host = "https://repository.de.dariah.eu/1.0/oaipmh/oai";
  // private static String host = "https://dhrepworkshop.de.dariah.eu/1.0/oaipmh/oai";
  private static String oaipmhEndpoint = "https://trep.de.dariah.eu/1.0/oaipmh/oai";
  // private static String host = "https://vm1.dariah.local/1.0/oaipmh/oai";

  // **
  // STATICS
  // **

  // Some JAXRS things.
  private static OAIPMHProducer JAXRSClient;
  private static Client oaipmhWebClient;

  // **
  // PREPARATIONS
  // **

  /**
   * @throws Exception
   */
  @BeforeClass
  public static void setUpBeforeClass() throws Exception {

    // Get OAI-PMH REST endpoint and HTTP client.
    System.out.println("Getting OAI-PMH HTTP client...");

    // Get proxy first, set policy.
    JAXRSClient = JAXRSClientFactory.create(oaipmhEndpoint, OAIPMHProducer.class);
    HTTPConduit conduit = WebClient.getConfig(JAXRSClient).getHttpConduit();
    HTTPClientPolicy policy = new HTTPClientPolicy();
    policy.setReceiveTimeout(OaiPmhTestUtilities.OAIPMH_CLIENT_TIMEOUT);
    conduit.setClient(policy);

    // Create Web Client from Web Proxy.
    oaipmhWebClient = WebClient.client(JAXRSClient);
  }

  /**
   * @throws Exception
   */
  @AfterClass
  public static void tearDownAfterClass() throws Exception {}

  /**
   * @throws Exception
   */
  @Before
  public void setUp() throws Exception {}

  /**
   * @throws Exception
   */
  @After
  public void tearDown() throws Exception {}

  // **
  // TESTS
  // **

  /**
   * @throws IOException
   */
  @Test
  public void testGetVersion() throws IOException {

    String shouldStartWith = "oaipmh-core";

    System.out.println(OaiPmhTestUtilities.TESTING + "#GETVERSION for '" + shouldStartWith + "'");

    String versionString = JAXRSClient.getVersion();

    if (!versionString.startsWith(shouldStartWith)) {
      System.err.println(
          OaiPmhTestUtilities.ERROR + ": response should start with '" + shouldStartWith + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + versionString);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testRootWithSlash() throws IOException {

    String url = "oai/";
    String shouldContain = "DARIAH-DE Repository";

    testRootURL(url, shouldContain);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testRootWithoutSlash() throws IOException {

    String url = "oai";
    String shouldContain = "DARIAH-DE Repository";

    testRootURL(url, shouldContain);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testBlubbVerb() throws IOException {

    String url = "verb=BLUBB";
    String shouldContain = "Illegal OAI verb";

    testRootURL(url, shouldContain);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testIdentify() throws IOException {

    String verb = "Identify";
    String shouldContain = "DARIAH-DE Repository";

    System.out.println(OaiPmhTestUtilities.TESTING + "#IDENTIFY");

    Response response = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, "verb=" + verb);
    int status = response.getStatus();

    String responseString = IOUtils.readStringFromStream((InputStream) response.getEntity());

    if (status != HttpStatus.SC_OK || !responseString.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + responseString);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListSets() throws IOException {

    // repository.de.dariah.eu
    String set = "hdl:21.11113/0000-000B-CADD-9";

    if (oaipmhEndpoint.contains("vm1")) {
      // VM1: <setSpec>hdl:21.T11991/0000-001B-4CE2-F</setSpec>
      // VM1: <setName>DUNGEON MASTER images</setName>
      set = "hdl:21.T11991/0000-001B-4CE2-F";
    } else if (oaipmhEndpoint.contains("trep")) {
      set = "hdl:21.T11991/0000-0008-C3AC-A";
    } else if (oaipmhEndpoint.contains("dhrepworkshop")) {
      set = "hdl:21.T11991/0000-0013-8FEE-F";
    }

    String expected = set;
    testListSet(set, expected);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsSinglePageDC() throws IOException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "#LISTRECORDS with expected single page (without token)");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    testList("ListRecords", "oai_dc", set);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsMorePagesDC() throws IOException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "#LISTRECORDS with expected resumption token usage");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    testList("ListRecords", "oai_dc", set);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsMorePagesDCFromUntil() throws IOException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "#LISTRECORDS with expected resumption token usage");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    // FIXME Values for vm1 only!
    String from = "2016-03-03";
    String until = "2018-05-04";
    int expectedPages = 5;

    int pageAmount = testList("ListRecords", "oai_dc", set, from, until);

    if (pageAmount != expectedPages) {
      assertTrue(false);
    }
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsSinglePageDATACITE() throws IOException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "#LISTRECORDS with expected single page (without token)");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    testList("ListRecords", "oai_datacite", set);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsMorePagesDATASITE() throws IOException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "#LISTRECORDS with expected resumption token usage");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    testList("ListRecords", "oai_datacite", set);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsMorePagesDATASITEFromUntil() throws IOException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "#LISTRECORDS with expected resumption token usage");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    // FIXME Values for vm1 only!
    String from = "2016-03-03";
    String until = "2018-05-04";
    int expectedPages = 5;

    int pageAmount = testList("ListRecords", "oai_datacite", set, from, until);

    if (pageAmount != expectedPages) {
      assertTrue(false);
    }
  }

  /**
   * @throws IOException
   */
  @Test
  @Ignore
  public void testGetRecordDC() throws IOException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "#GETRECORD with expected single page (without token)");

    String identifier = "";
    String shouldContain = "";

    String path = "verb=GetRecord&identifier=" + identifier + "&metadataPrefix=oai_dc";

    // FIXME Complete test!
  }

  /**
   * @throws IOException
   */
  @Test
  @Ignore
  public void testGetRecordDATASITE() throws IOException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "#GETRECORD with expected single page (without token)");

    String identifier = "";
    String shouldContain = "";

    String path = "verb=GetRecord&identifier=" + identifier + "&metadataPrefix=oai_datacite";

    // FIXME Complete test!
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifierSinglePageDC() throws IOException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "#LISTIDENTIFIERS with expected single page (without token)");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    testList("ListIdentifiers", "oai_dc", set);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersMorePagesDC() throws IOException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "#LISTIDENTIFIERS with expected resumption token usage");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    testList("ListIdentifiers", "oai_dc", set);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifierSinglePageDATACITE() throws IOException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "#LISTIDENTIFIERS with expected single page (without token)");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    testList("ListIdentifiers", "oai_datacite", set);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersMorePagesDATASITE() throws IOException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "#LISTIDENTIFIERS with expected resumption token usage");

    // Testing set: Bahnsen 1670_Erschließung_20161017.xlsx
    // String set = "21.11113/0000-000B-CAD1-5";
    String set = "";

    testList("ListIdentifiers", "oai_datacite", set);
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * TODO Generalise in OaiPmhTestUtilities!
   * 
   * @param theResponse
   * @return Resumption token string, if existing, -1 if either no <resumptionToken> tag is
   *         existing, tag is existing and has no token value.
   * @throws IOException
   */
  private static String examineResumptionTokenTag(Response theResponse, String recordOrHeader,
      String oldtok) throws IOException {

    String responseString = IOUtils.readStringFromStream((InputStream) theResponse.getEntity());

    // Test for OAIPMH errors.
    if (responseString.contains("<error code=\"badArgument\">")) {
      System.err.println(OaiPmhTestUtilities.ERROR + " IN OAIPMH RESPONSE: " + responseString);
      assertTrue(false);
    }

    // Count response objects at first.
    int recordCount = 0;
    int i = responseString.indexOf("<" + recordOrHeader + ">", 0);
    while (i != -1) {
      recordCount++;
      i++;
      i = responseString.indexOf("<" + recordOrHeader + ">", i);
    }

    System.out.println("\t" + recordOrHeader + "s: " + recordCount);

    // Check if token tag is existing.
    int tokStart = responseString.indexOf("<resumptionToken");
    int tokEnd = responseString.indexOf("</resumptionToken");

    if (tokStart == -1 && tokEnd == -1) {
      System.out.println("\ttoken: no token");

      return "-1";
    }

    String restokTmp = responseString.substring(tokStart, tokEnd);
    // Get token tag.
    String toktag = restokTmp.substring(0, restokTmp.indexOf(">") + 1).trim();
    System.out.println("\ttokentag: " + toktag);

    // Get token.
    String restok = restokTmp.substring(restokTmp.indexOf(">") + 1).trim();
    System.out.println("\ttoken: " + restok);

    // Check if old and new token are equal or not.
    boolean tokchanged = !oldtok.equals(restok);
    System.out.println("\ttokchngd: " + tokchanged);

    // Get completeListSize and cursor.
    String sizeStr = toktag.substring(toktag.indexOf("completeListSize=\"") + 18);
    int size = Integer.parseInt(sizeStr.substring(0, sizeStr.indexOf("\"")));
    String cursorStr = toktag.substring(toktag.indexOf("cursor=\"") + 8);
    int cursor = Integer.parseInt(cursorStr.substring(0, cursorStr.indexOf("\"")));
    System.out.println("\tsize: " + size + " / " + cursor);

    // If token is provided, and we have less than 100 elements: mekkern!
    if (!restok.isEmpty()) {
      synchronized (OaiPmhDariahdeOnlineTests.class) {
        // Check <record> or <header> count, must be 100!
        if (recordCount != 100) {
          System.err.println(OaiPmhTestUtilities.ERROR + ": " + recordOrHeader
              + " count mismatch, must be 100 if token is provided, but is " + recordCount);
        }
        if (size <= 100) {
          System.err.println(OaiPmhTestUtilities.ERROR
              + ": completeListSize count mismatch, must be > 100 if token is provided, but is "
              + size);
        }
        if (recordCount != 100 || size <= 100) {
          assertTrue(false);
        }
      }
    }

    // If no token is provided, stop querying.
    else {
      // Check <record> count, must be completeListSize % 100.
      if (recordCount != size % 100) {
        System.err.println(
            OaiPmhTestUtilities.ERROR + ": " + recordOrHeader + " count mismatch, should be "
                + size % 100 + ", but is " + recordCount);
        assertTrue(false);
      }

      // No resumption token available in response.
      return "-1";
    }

    return restok;
  }

  /**
   * @param theVerb
   * @param theMetadataPrefix
   * @param theSet
   * @return
   * @throws IOException
   */
  private static int testList(String theVerb, String theMetadataPrefix, String theSet)
      throws IOException {
    return testList(theVerb, theMetadataPrefix, theSet, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL);
  }

  /**
   * TODO Generalise in OaiPmhTestUtilities!
   * 
   * @param theVerb
   * @param theMetadataPrefix
   * @param theSet
   * @param from
   * @param until
   * @return How many pages were delivered
   * @throws IOException
   */
  private static int testList(String theVerb, String theMetadataPrefix, String theSet, String from,
      String until) throws IOException {

    int result;

    long startTime = System.currentTimeMillis();

    String testOccurance = "header";
    if (theVerb.equals(OaiPmhTestUtilities.VERB_LIST_RECORDS)) {
      testOccurance = "record";
    }

    String path = "verb=" + theVerb + "&metadataPrefix=" + theMetadataPrefix;

    if (theSet != null && !theSet.isEmpty()) {
      path += "&set=" + theSet;
    }

    if (from != null && until != null) {
      path += "&from=" + from + "&until=" + until;
    }

    Response response = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path);
    int status = response.getStatus();

    long timeRunning = System.currentTimeMillis() - startTime;
    System.out.println("\ttime: " + OaiPmhTestUtilities.getDurationInSecs(timeRunning));

    String restok = examineResumptionTokenTag(response, testOccurance, "");

    result = 1;
    while (status == HttpStatus.SC_OK && !restok.equals("-1")) {
      path = "verb=" + theVerb + "&resumptionToken=" + restok;
      response = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path);
      timeRunning = System.currentTimeMillis() - startTime;
      System.out.println("\ttime: " + OaiPmhTestUtilities.getDurationInSecs(timeRunning));
      restok = examineResumptionTokenTag(response, testOccurance, restok);
      result++;
    }

    System.out.println("\tpage amount: " + result);
    System.out.println(OaiPmhTestUtilities.OK);

    return result;
  }

  /**
   * @param theSet
   * @throws IOException
   */
  private static void testListSet(String theSet, String theExpectedResponse) throws IOException {

    String verb = "ListSets";

    System.out.println(OaiPmhTestUtilities.TESTING + "#LISTSETS");

    Response response = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, "verb=" + verb);
    int status = response.getStatus();

    String responseString = IOUtils.readStringFromStream((InputStream) response.getEntity());

    if (status != HttpStatus.SC_OK || !responseString.contains(theExpectedResponse)) {
      System.err.println("\tstatus: " + status);
      System.err.println(
          OaiPmhTestUtilities.ERROR + ": response should contain '" + theExpectedResponse + "'");
      System.err.println(responseString);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + responseString);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @param theURL
   * @param shouldContain
   * @throws IOException
   */
  private static void testRootURL(String theURL, String shouldContain) throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "#ROOTURL");

    Response response = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, theURL);
    int status = response.getStatus();

    String responseString = IOUtils.readStringFromStream((InputStream) response.getEntity());

    if (status != HttpStatus.SC_OK || !responseString.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + responseString);
    System.out.println(OaiPmhTestUtilities.OK);
  }

}
