package info.textgrid.middleware;

import static org.junit.Assert.assertTrue;
import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.Callable;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import javax.ws.rs.core.Response;
import org.apache.cxf.helpers.IOUtils;
import org.apache.cxf.jaxrs.client.Client;
import org.apache.cxf.jaxrs.client.JAXRSClientFactory;
import org.apache.cxf.jaxrs.client.WebClient;
import org.apache.cxf.transport.http.HTTPConduit;
import org.apache.cxf.transports.http.configuration.HTTPClientPolicy;
import org.apache.http.HttpStatus;
import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.Ignore;
import org.junit.Test;

/**
 * <p>
 * Some online tests for the TextGrid OAIMPH service.
 * </p>
 * 
 * @author Stefan E. Funk, SUB Göttingen
 */
@Ignore
public class OaiPmhTextgridOnlineTests {

  // TODO Configure config files for all the different OAI-PMH service instances!

  // **
  // The OAIPMH host to be tested.
  // **

  // private static String host = "https://textgridlab.org/1.0/tgoaipmh/oai";
  // private static String oaipmhEndpoint = "https://test.textgridlab.org/1.0/tgoaipmh/oai";
  private static String oaipmhEndpoint = "https://dev.textgridlab.org/1.0/tgoaipmh/oai";

  // NOTE Test with "mvn jetty:run" in oaipmh-webapp module folder and SSH tunneling "ssh -L
  // 9202:localhost:9202 -l [username] dev.textgridlab.org", please use original OAIPMH config file
  // from /etc/dhrep/oaipmh/oaipmh.textgrid.properties and copy it to
  // /etc/dhrep/oaipmh/oaipmh.config on your local machine!
  // private static String oaipmhEndpoint = "http://localhost:8095/oaipmh/oai";
  // NOTE Use "mvn tomcat:run" here!
  // private static String oaipmhEndpoint = "http://localhost:8080/oaipmh-webapp/oai";

  // **
  // FINALS
  // **

  private static final String OAIDC_SCHEMA_FORMAT = "oai_dc";
  private static final String IDIOMMETS_SCHEMA_FORMAT = "oai_idiom_mets";
  private static final String DATACITE_SCHEMA_FORMAT = "oai_datacite";
  private static final String EXPECTED_OAIDC_FORMAT_CONTENT = "<oai_dc:dc>";
  private static final String EXPECTED_IDIOMMETS_FORMAT_CONTENT = "<mets ";
  private static final String EXPECTED_DATACITE_FORMAT_CONTENT = "<datacite:resource>";

  // **
  // STATICS
  // **

  // Some JAXRS things.
  private static OAIPMHProducer JAXRSClient;
  private static Client oaipmhWebClient;

  // **
  // PREPARATIONS
  // **

  /**
   * @throws Exception
   */
  @BeforeClass
  public static void setUpBeforeClass() throws Exception {

    // Get OAI-PMH REST endpoint and HTTP client.
    System.out.println("Getting OAI-PMH HTTP client --> " + oaipmhEndpoint + " <--");

    // Get proxy first, set policy.
    JAXRSClient = JAXRSClientFactory.create(oaipmhEndpoint, OAIPMHProducer.class);
    HTTPConduit conduit = WebClient.getConfig(JAXRSClient).getHttpConduit();
    HTTPClientPolicy policy = new HTTPClientPolicy();
    policy.setReceiveTimeout(OaiPmhTestUtilities.OAIPMH_CLIENT_TIMEOUT);
    conduit.setClient(policy);

    // Create Web Client from Web Proxy.
    oaipmhWebClient = WebClient.client(JAXRSClient);
  }

  /**
   * @throws Exception
   */
  @AfterClass
  public static void tearDownAfterClass() throws Exception {}

  /**
   * @throws Exception
   */
  @Before
  public void setUp() throws Exception {}

  /**
   * @throws Exception
   */
  @After
  public void tearDown() throws Exception {}

  // **
  // TESTS
  // **

  /**
   * @throws IOException
   */
  @Test
  public void testGetVersion() throws IOException {

    String shouldStartWith = "oaipmh-core";

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetVersion()");

    String versionString = JAXRSClient.getVersion();

    if (!versionString.startsWith(shouldStartWith)) {
      System.err.println(
          OaiPmhTestUtilities.ERROR + ": response should start with '" + shouldStartWith + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + versionString);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testRootUrl() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testRootUrl()");

    String shouldContain = "textgridrep.org";

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, "oai",
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testIdentify() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testIdentify()");

    String shouldContain = "textgridrep.org";

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient,
        "verb=" + OaiPmhTestUtilities.VERB_IDENTIFY, OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListMetadataFormats() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListMetadataFormats()");

    String shouldContainOaidc = "oai_dc";
    String shouldContainOaiidiommets = "oai_idiom_mets";

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient,
        "verb=" + OaiPmhTestUtilities.VERB_LIST_METADATA_FORMATS,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContainOaidc)
        || !response.contains(shouldContainOaiidiommets)) {
      System.err.println("\tstatus: " + status);
      System.err.println(
          OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContainOaidc + "' and '"
              + shouldContainOaiidiommets + "'");
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListSets() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListSets()");

    // Testing for set: TextGrid Digitale Bibliothek.
    String shouldContainTEXTGRIDLABORG = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";
    String shouldContainDEFTEXTGRIDLABORG = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient,
        "verb=" + OaiPmhTestUtilities.VERB_LIST_SETS, OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContainTEXTGRIDLABORG)
        && !response.contains(shouldContainDEFTEXTGRIDLABORG)) {
      System.err.println("\tstatus: " + status);
      System.err.println(OaiPmhTestUtilities.ERROR + ": response should contain (textgridlab.org) '"
          + shouldContainTEXTGRIDLABORG + "' or (dev.textgridlab.org) '"
          + shouldContainDEFTEXTGRIDLABORG + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordOAIDC() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordOAIDC()");

    String identifier = "textgrid:vqn0.0";
    String shouldContain = "Heidi";

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + identifier + "&metadataPrefix=" + OaiPmhTestUtilities.OAI_DC_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordOAIDCNoPrefix() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordOAIDCNoPrefix()");

    String identifier = "mq05.0";
    String shouldContain = "tvitt@textgrid.de";

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + identifier + "&metadataPrefix=" + OaiPmhTestUtilities.OAI_DC_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordIDIOMMETS() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordIDIOMMETS()");

    String identifier = "textgrid:2sg18.0";
    String shouldContain = "Aguateca";

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + identifier + "&metadataPrefix=" + OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordIDIOMMETSNoPrefix() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordIDIOMMETSNoPrefix()");

    String identifier = "2sg18.0";
    String shouldContain = "Aguateca";

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + identifier + "&metadataPrefix=" + OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordDATACITE() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordDATACITE()");

    String identifier = "textgrid:mq05.0";
    String shouldContain = "Spraach-en Höfflichkeit";

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier="
        + identifier + "&metadataPrefix=" + OaiPmhTestUtilities.OAI_DATACITE_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();

    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testGetRecordDATACITENoPrefix() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testGetRecordDATACITENoPrefix()");

    String identifier = "mq05.0";
    String shouldContain = "Spraach-en Höfflichkeit";

    String path = "verb=" + OaiPmhTestUtilities.VERB_GET_RECORD + "&identifier=" + identifier
        + "&metadataPrefix=" + OaiPmhTestUtilities.OAI_DATACITE_PREFIX;

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, path,
        OaiPmhTestUtilities.NO_THREAD_NAME);
    int status = httpResponse.getStatus();
    String response = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    if (status != HttpStatus.SC_OK || !response.contains(shouldContain)) {
      System.err.println("\tstatus: " + status);
      System.err
          .println(OaiPmhTestUtilities.ERROR + ": response should contain '" + shouldContain + "'");
      System.err.println(response);
      assertTrue(false);
    }

    System.out.println("\tresponse: " + response);
    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsWithSetOAIDCSinglePage() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsOAIDCSinglePage()");

    // IIIF
    // String set = "project:TGPR-2e532567-c3e3-36f5-3ecc-520cea7904cc";
    // ELTEX Test
    String set = "project:TGPR-59722e29-efcb-330f-b9b4-5ef2fb08edab";

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, set,
        OaiPmhTestUtilities.OAI_DC_PREFIX, 0, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsOAIDCMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsOAIDCMorePages()");

    // Testing set: Digitale Bibliothek (on prod and dev available only...)
    String set = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, set,
        OaiPmhTestUtilities.OAI_DC_PREFIX, 20, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsOAIDCMorePagesFromUntil() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsOAIDCMorePagesFromUntil()");

    // Testing set: Digitale Bibliothek (on prod and dev available only...)
    String set = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";
    String from = "2012-01-04T01:00:00";
    String until = "2012-01-04T12:00:00";

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, set,
        OaiPmhTestUtilities.OAI_DC_PREFIX, 0, 100, OaiPmhTestUtilities.NO_THREAD_NAME, from, until,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test(expected = IOException.class)
  public void testListRecordsIDIOMMETSSomePagesAndMetadataFormatWithRestok() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING
        + "testListRecordsIDIOMMETSSomePagesAndMetadataFormatWithRestok()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS,
        OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 3, 30, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  @Ignore
  public void testListRecordsIDIOMMETSAllPages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsIDIOMMETSAllPages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS,
        OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 0, 30, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsIDIOMMETSMorePagesFromUntil() throws IOException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testListRecordsIDIOMMETSMorePagesFromUntil()");

    String from = "2015-11-12T01:00:00";
    String until = "2016-03-02T12:00:00";

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS,
        OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 6, 30, OaiPmhTestUtilities.NO_THREAD_NAME, from,
        until, OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsIDIOMMETSMorePages() throws IOException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testListRecordsIDIOMMETSMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS, OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 6, 30, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsDATACITEMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsDATACITEMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS,
        OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 15, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  @Ignore
  public void testListRecordsDATACITEAllPages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsDATACITEAllPages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS,
        OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 0, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListRecordsDATACITEMorePagesFromUntil() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListRecordsDATACITEMorePagesFromUntil()");

    String from = "2012-01-04T01:00:00";
    String until = "2012-01-04T12:00:00";

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_RECORDS,
        OaiPmhTestUtilities.NO_SET,
        OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 15, 100, OaiPmhTestUtilities.NO_THREAD_NAME, from,
        until, OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifierOAIDCSinglePage() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListIdentifierOAIDCSinglePage()");

    // IIIF
    // String set = "project:TGPR-2e532567-c3e3-36f5-3ecc-520cea7904cc";
    // ELTEX Test
    String set = "project:TGPR-59722e29-efcb-330f-b9b4-5ef2fb08edab";

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS, set,
        OaiPmhTestUtilities.OAI_DC_PREFIX, 0, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersOAIDCMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListIdentifiersOAIDCMorePages()");

    // Testing set: Digitale Bibliothek (on prod and dev available only...)
    String set = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS, set,
        OaiPmhTestUtilities.OAI_DC_PREFIX, 30, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test(expected = IOException.class)
  public void testListIdentifiersOAIDCMorePagesAndMetadataFormatWithRestok() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING
        + "testListIdentifiersOAIDCMorePagesAndMetadataFormatWithRestok()");

    // Testing set: Digitale Bibliothek (on prod and dev available only...)
    String set = "project:TGPR-372fe6dc-57f2-6cd4-01b5-2c4bbefcfd3c";

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS, set,
        OaiPmhTestUtilities.OAI_DC_PREFIX, 30, 100, OaiPmhTestUtilities.NO_THREAD_NAME,
        OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
        OaiPmhTestUtilities.METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  @Ignore
  public void testListIdentifiersIDIOMMETSAllPages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListIdentifiersIDIOMMETSAllPages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 0, 30,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersIDIOMMETSMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListIdentifiersIDIOMMETSMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 6, 30,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test(expected = IOException.class)
  public void testListIdentifiersIDIOMMETSSomePagesAndMetadataFormatWithRestok()
      throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING
        + "testListIdentifiersIDIOMMETSSomePagesAndMetadataFormatWithRestok()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 3, 30,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   */
  @Test
  public void testListIdentifiersDATACITEMorePages() throws IOException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testListIdentifiersDATACITEMorePages()");

    testList(oaipmhWebClient, OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
        OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 22, 100,
        OaiPmhTestUtilities.NO_THREAD_NAME, OaiPmhTestUtilities.NO_FROM,
        OaiPmhTestUtilities.NO_UNTIL, OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);

    System.out.println(OaiPmhTestUtilities.OK);
  }

  /**
   * @throws IOException
   * @throws ExecutionException
   * @throws InterruptedException
   */
  @Test
  public void testListIdentifiersConcurrentlyIDIOMMETSMorePages()
      throws IOException, InterruptedException, ExecutionException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "testListIdentifiersConcurrentlyIDIOMMETSMorePages()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 3, 30, "C1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 5, 30, "C2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 8, 30, "C2"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [C1]=" + f1.get() + ", [C2]=" + f2.get()
        + ", [C3]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws ExecutionException
   * @throws InterruptedException
   */
  @Test
  public void testListRecordsConcurrentlyIDIOMMETSMorePages()
      throws IOException, InterruptedException, ExecutionException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testListRecordsConcurrentlyIDIOMMETSMorePages()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 2, 30, "D1"));
    Future<Boolean> f2 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 3, 30, "D2"));
    Future<Boolean> f3 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 4, 30, "D3"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [D1]=" + f1.get() + ", [D2]=" + f2.get()
        + ", [D3]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiersDC()
      throws InterruptedException, ExecutionException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListIdentifiersDC()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 36, 100, "A1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 27, 100, "A2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 19, 100, "A3"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [A1]=" + f1.get() + ", [A2]=" + f2.get()
        + ", [A3]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiersIDIOM()
      throws InterruptedException, ExecutionException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListIdentifiersIDIOM()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 13, 30, "A1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 35, 30, "A2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, 11, 30, "A3"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [A1]=" + f1.get() + ", [A2]=" + f2.get()
        + ", [A3]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiersDATACITE()
      throws InterruptedException, ExecutionException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListIdentifiersDATACITE()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 13, 100, "A1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 20, 100, "A2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DATACITE_PREFIX, 8, 100, "A3"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [A1]=" + f1.get() + ", [A2]=" + f2.get()
        + ", [A3]=" + f3.get());
  }

  /**
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListIdentifiersDCAndIDIOMMets()
      throws InterruptedException, ExecutionException {

    System.out.println(
        OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListIdentifiersDCAndIDIOMMets()");

    int idiomPagesToTest = 10;
    int idiomRecordsPerPage = 30;
    if (!oaipmhEndpoint.contains("dev.textgridlab.org")) {
      idiomPagesToTest = 1;
      idiomRecordsPerPage = 1;
    }

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 53, 100, "DC1"));
    Future<Boolean> f2 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 28, 100, "DC2"));
    Future<Boolean> f3 = executor
        .submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_IDENTIFIERS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, idiomPagesToTest,
            idiomRecordsPerPage, "IDIOM"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [DC1]=" + f1.get() + ", [DC2]=" + f2.get()
        + ", [IDIOM]=" + f3.get());
  }

  /**
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListRecordsDC()
      throws InterruptedException, ExecutionException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListRecordsDC()");

    ExecutorService executor = Executors.newFixedThreadPool(4);

    Future<Boolean> f1 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 44, 100, "DC1"));
    Future<Boolean> f2 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 33, 100, "DC2"));
    Future<Boolean> f3 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 66, 100, "DC3"));
    Future<Boolean> f4 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 28, 100, "DC4"));

    executor.shutdown();

    System.out.println(
        OaiPmhTestUtilities.OK + ": [DC1]=" + f1.get() + ", [DC2]=" + f2.get() + ", [DC3]="
            + f3.get() + ", [DC4]=" + f4.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testRestokConcurrentlyListRecords() throws InterruptedException, ExecutionException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testRestokConcurrentlyListRecords()");

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 6, 100, "B1"));
    Future<Boolean> f2 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 7, 100, "B2"));
    Future<Boolean> f3 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 8, 100, "B3"));
    Future<Boolean> f4 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 9, 100, "B4"));

    executor.shutdown();

    System.out.println(
        OaiPmhTestUtilities.OK + ": [B1]=" + f1.get() + ", [B2]=" + f2.get() + ", [B3]=" + f3.get()
            + ", [B4]=" + f4.get());
  }

  /**
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testListRecordsRestokConcurrentlyDCAndIDIOMMets()
      throws InterruptedException, ExecutionException {

    System.out
        .println(OaiPmhTestUtilities.TESTING + "testListRecordsRestokConcurrentlyDCAndIDIOMMets()");

    int idiomPagesToTest = 9;
    int idiomRecordsPerPage = 30;
    if (!oaipmhEndpoint.contains("dev.textgridlab.org")) {
      idiomPagesToTest = 1;
      idiomRecordsPerPage = 1;
    }

    ExecutorService executor = Executors.newFixedThreadPool(3);

    Future<Boolean> f1 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 34, 100, "DC1"));
    Future<Boolean> f2 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_DC_PREFIX, 27, 100, "DC2"));
    Future<Boolean> f3 =
        executor.submit(new OaiPmhResumptionTokenTestThread(OaiPmhTestUtilities.VERB_LIST_RECORDS,
            OaiPmhTestUtilities.NO_SET, OaiPmhTestUtilities.OAI_IDIOMMETS_PREFIX, idiomPagesToTest,
            idiomRecordsPerPage, "IDIOM"));

    executor.shutdown();

    System.out.println(OaiPmhTestUtilities.OK + ": [DC1]=" + f1.get() + ", [DC2]=" + f2.get()
        + ", [IOIOM]=" + f3.get());
  }

  /**
   * @throws IOException
   * @throws InterruptedException
   * @throws ExecutionException
   */
  @Test
  public void testConcurrentlyOAIDCGetRecord() throws InterruptedException, ExecutionException {

    System.out.println(OaiPmhTestUtilities.TESTING + "testConcurrentlyOAIDCGetRecord()");

    String id1 = "textgrid:mq05.0";
    String id2 = "textgrid:jgv6.0";
    String id3 = "textgrid:w7rz.0";
    String id4 = "textgrid:w36b.0";
    String id5 = "textgrid:mcdv.0";
    String id6 = "textgrid:v0qx.0";
    String id7 = "textgrid:xkck.0";
    String id8 = "textgrid:t3m1.0";
    String id9 = "textgrid:10rsq.0";
    String id10 = "textgrid:mq05.0";
    String id11 = "textgrid:jgv6.0";
    String id12 = "textgrid:w7rz.0";
    String id13 = "textgrid:w36b.0";
    String id14 = "textgrid:mcdv.0";
    String id15 = "textgrid:v0qx.0";
    String id16 = "textgrid:xkck.0";
    String id17 = "textgrid:t3m1.0";
    String id18 = "textgrid:10rsq.0";

    if (!oaipmhEndpoint.contains("dev.textgridlab.org")) {

      ExecutorService executor = Executors.newFixedThreadPool(18);

      Future<Boolean> f1 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id1,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR1"));
      Future<Boolean> f2 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id2,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR2"));
      Future<Boolean> f3 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id3,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR3"));
      Future<Boolean> f4 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id4,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR4"));
      Future<Boolean> f5 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id5,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR5"));
      Future<Boolean> f6 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id6,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR6"));
      Future<Boolean> f7 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id7,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR7"));
      Future<Boolean> f8 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id8,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR8"));
      Future<Boolean> f9 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id9,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR9"));
      Future<Boolean> f10 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id10,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR10"));
      Future<Boolean> f11 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id11,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR11"));
      Future<Boolean> f12 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id12,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR12"));
      Future<Boolean> f13 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id13,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR13"));
      Future<Boolean> f14 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id14,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR14"));
      Future<Boolean> f15 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id15,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR15"));
      Future<Boolean> f16 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id16,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR16"));
      Future<Boolean> f17 = executor
          .submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id17,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR17"));
      Future<Boolean> f18 =
          executor.submit(new OaiPmhGetRecordTestThread(OaiPmhTestUtilities.VERB_GET_RECORD, id18,
              OaiPmhTestUtilities.OAI_DC_PREFIX, "GR18"));

      executor.shutdown();

      System.out.println(OaiPmhTestUtilities.OK + "[GR1]=" + f1.get() + ", [GR2]=" + f2.get()
          + ", [GR3]=" + f3.get() + ", [GR4]=" + f4.get() + ", [GR5]=" + f5.get() + ", [GR6]="
          + f6.get() + ", [GR7]=" + f7.get() + ", [GR8]=" + f8.get() + ", [GR9]=" + f9.get()
          + ", [GR10]=" + f10.get() + ", [GR11]=" + f11.get() + ", [GR12]=" + f12.get()
          + ", [GR13]=" + f13.get() + ", [GR14]=" + f14.get() + ", [GR15]=" + f15.get()
          + ", [GR16]=" + f16.get() + ", [GR17]=" + f17.get() + ", [GR18]=" + f18.get());
    } else {
      System.out.println(OaiPmhTestUtilities.NOT_TESTED);
    }
  }

  /**
   * <p>
   * Starts a new OAI-PMH thread.
   * </p>
   */

  private class OaiPmhResumptionTokenTestThread implements Callable<Boolean> {

    protected String verb;
    protected String set;
    protected String metadataPrefix;
    protected int numberOfPagesToTest;
    protected int recordsExpectedPerRequest;
    protected String threadName;

    /**
     * @param theVerb
     * @param theSet
     * @param theMetadataPrefix
     * @param numberOfPagesToTest
     * @param recordsExpectedPerRequest
     */
    public OaiPmhResumptionTokenTestThread(String theVerb, String theSet, String theMetadataPrefix,
        int numberOfPagesToTest, int recordsExpectedPerRequest, String theThreadName) {
      this.verb = theVerb;
      this.set = theSet;
      this.metadataPrefix = theMetadataPrefix;
      this.numberOfPagesToTest = numberOfPagesToTest;
      this.recordsExpectedPerRequest = recordsExpectedPerRequest;
      this.threadName = theThreadName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.lang.Thread#run()
     */
    public Boolean call() throws IOException {
      testList(oaipmhWebClient, this.verb, this.set, this.metadataPrefix,
          this.numberOfPagesToTest, this.recordsExpectedPerRequest, this.threadName,
          OaiPmhTestUtilities.NO_FROM, OaiPmhTestUtilities.NO_UNTIL,
          OaiPmhTestUtilities.NO_METADATA_FORMAT_WITH_RESTOK);
      return true;
    }
  }

  /**
   * <p>
   * Starts a new OAI-PMH thread.
   * </p>
   */

  private class OaiPmhGetRecordTestThread implements Callable<Boolean> {

    protected String verb;
    protected String identifier;
    protected String metadataPrefix;
    protected String threadName;

    /**
     * @param theVerb
     * @param theIdentifier
     * @param theMetadataPrefix
     * @param theThreadName
     */
    public OaiPmhGetRecordTestThread(String theVerb, String theIdentifier, String theMetadataPrefix,
        String theThreadName) {
      this.verb = theVerb;
      this.identifier = theIdentifier;
      this.metadataPrefix = theMetadataPrefix;
      this.threadName = theThreadName;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.util.concurrent.Callable#call()
     */
    public Boolean call() throws IOException {

      String url = "verb=" + this.verb + "&identifier=" + this.identifier + "&metadataPrefix="
          + this.metadataPrefix;
      OaiPmhTestUtilities.getHttpResponse(oaipmhWebClient, url, this.threadName);

      return true;
    }
  }

  // **
  // PRIVATE METHODS
  // **

  /**
   * TODO Generalise in OaiPmhTestUtilities!
   * 
   * @param theClient
   * @param theVerb
   * @param theSet
   * @param theMetadataPrefix
   * @param maxNumberOfPagesToTest
   * @param recordsExpectedPerRequest
   * @param theThreadName
   * @param from
   * @param until
   * @param resumptionTokenANDMetadataPrefix
   * @throws IOException
   */
  protected static void testList(Client theClient, final String theVerb, final String theSet,
      final String theMetadataPrefix, final int maxNumberOfPagesToTest,
      final int recordsExpectedPerRequest, String theThreadName, final String from,
      final String until, boolean resumptionTokenANDMetadataPrefix) throws IOException {

    theThreadName = tn(theThreadName);

    long startTime = System.currentTimeMillis();

    String testOccurance = "header";
    if (theVerb.equals(OaiPmhTestUtilities.VERB_LIST_RECORDS)) {
      testOccurance = "record";
    }

    String path = "verb=" + theVerb + "&metadataPrefix=" + theMetadataPrefix;

    if (theSet != null) {
      path += "&set=" + theSet;
    }

    if (from != null && until != null) {
      path += "&from=" + from + "&until=" + until;
    }

    Response httpResponse = OaiPmhTestUtilities.getHttpResponse(theClient, path, theThreadName);
    int status = httpResponse.getStatus();

    int loopCount = 1;
    long timeRunning = System.currentTimeMillis() - startTime;
    System.out.println("\t" + theThreadName + "time: "
        + OaiPmhTestUtilities.getDurationInSecs(timeRunning) + " (loop " + loopCount
        + (maxNumberOfPagesToTest != 0 ? "/" + maxNumberOfPagesToTest : "") + ")");

    String responseString = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());

    // Test resumption token tags.
    String restok = examineResumptionTokenTag(responseString, testOccurance,
        OaiPmhTestUtilities.NO_TOKEN, recordsExpectedPerRequest, loopCount, theThreadName);

    // Test general metadata content (if verb is listRecords!), must go conform with the metadata
    // prefix setting.
    if (theVerb.equals(OaiPmhTestUtilities.VERB_LIST_RECORDS)) {
      examineContent(responseString, theMetadataPrefix);
    }

    while (status == HttpStatus.SC_OK && !restok.equals(OaiPmhTestUtilities.NO_TOKEN)) {
      loopCount += 1;
      synchronized (theThreadName) {
        if (maxNumberOfPagesToTest > 0 && loopCount > maxNumberOfPagesToTest) {
          System.out
              .println("\t" + theThreadName + "TESTING ONLY " + maxNumberOfPagesToTest + " pages!");
          break;
        }
        path = "verb=" + theVerb + "&resumptionToken=" + restok;
        if (resumptionTokenANDMetadataPrefix) {
          path += "&metadataPrefix=" + theMetadataPrefix;
        }
        httpResponse = OaiPmhTestUtilities.getHttpResponse(theClient, path, theThreadName);
        timeRunning = System.currentTimeMillis() - startTime;
        System.out.println("\t" + theThreadName + "time: "
            + OaiPmhTestUtilities.getDurationInSecs(timeRunning) + " (loop " + loopCount
            + (maxNumberOfPagesToTest != 0 ? "/" + maxNumberOfPagesToTest : "") + ")");

        // Test resumption token tags.
        responseString = IOUtils.readStringFromStream((InputStream) httpResponse.getEntity());
        restok = examineResumptionTokenTag(responseString, testOccurance, restok,
            recordsExpectedPerRequest, loopCount, theThreadName);

        // Test general metadata content (if verb is listRecords!), must go conform with the
        // metadata // prefix setting.
        if (theVerb.equals(OaiPmhTestUtilities.VERB_LIST_RECORDS)) {
          examineContent(responseString, theMetadataPrefix);
        }
      }
    }

    // Only check for max loops, if we do have less loops and no resumption token, it will be fine,
    // too!
    if (!restok.equals(OaiPmhTestUtilities.NO_TOKEN) && loopCount < maxNumberOfPagesToTest) {
      System.out.println(theThreadName + OaiPmhTestUtilities.ERROR + ": Must have done "
          + maxNumberOfPagesToTest + " loops, but did only " + loopCount);
      assertTrue(false);
    }

    // Check for single loop, we need MORE!
    if (loopCount == 1) {
      System.out
          .println(theThreadName + OaiPmhTestUtilities.ERROR + ": Must have more than one loop!");
      assertTrue(false);
    }

    System.out.println("\t" + theThreadName + OaiPmhTestUtilities.OK);
  }

  /**
   * TODO Generalise in OaiPmhTestUtilities!
   * 
   * @param theResponse
   * @param recordOrHeader
   * @param oldtok
   * @param recordsExpectedPerRequest
   * @param loopCount
   * @param theThreadName
   * @return Resumption token string, if existing, -1 if either no <resumptionToken> tag is
   *         existing, tag is existing and has no token value.
   * @throws IOException
   */
  public static String examineResumptionTokenTag(final String theResponseString,
      final String recordOrHeader, final String oldtok, final int recordsExpectedPerRequest,
      final int loopCount, final String theThreadName) throws IOException {

    // Test for OAIPMH errors.
    if (theResponseString.contains("<error code=\"badArgument\">")) {
      String message =
          theThreadName + OaiPmhTestUtilities.ERROR + " IN OAIPMH RESPONSE: " + theResponseString;
      System.out.println(message);
      throw new IOException(message);
    }

    // Count response objects at first.
    int recordCount = 0;
    int i = theResponseString.indexOf("<" + recordOrHeader + ">", 0);
    while (i != -1) {
      recordCount++;
      i++;
      i = theResponseString.indexOf("<" + recordOrHeader + ">", i);
    }

    System.out.println("\t" + theThreadName + recordOrHeader + "s: " + recordCount);

    // Check if token tag is existing.
    int tokStart = theResponseString.indexOf("<resumptionToken");
    int tokEnd = theResponseString.indexOf("</resumptionToken");

    if (tokStart == -1 && tokEnd == -1) {
      System.out.println("\t" + theThreadName + "token: no token");

      return OaiPmhTestUtilities.NO_TOKEN;
    }

    String restokTmp = theResponseString.substring(tokStart, tokEnd);
    // Get token tag.
    String toktag = restokTmp.substring(0, restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + "tokentag: " + toktag);

    // Get token.
    String restok = restokTmp.substring(restokTmp.indexOf(">") + 1).trim();
    System.out.println("\t" + theThreadName + "token: " + restok);

    // Check if old and new token are equal or not.
    boolean tokchanged = !oldtok.equals(restok);
    System.out.println("\t" + theThreadName + "tokchngd: " + tokchanged);

    // Get completeListSize and cursor.
    String sizeStr = toktag.substring(toktag.indexOf("completeListSize=\"") + 18);
    int size = Integer.parseInt(sizeStr.substring(0, sizeStr.indexOf("\"")));
    String cursorStr = toktag.substring(toktag.indexOf("cursor=\"") + 8);
    int cursor = Integer.parseInt(cursorStr.substring(0, cursorStr.indexOf("\"")));
    System.out.println("\t" + theThreadName + "size: " + size + " / " + cursor);

    // If token is provided, test cursor and element count!
    if (!restok.isEmpty()) {
      synchronized (OaiPmhTextgridOnlineTests.class) {
        // Check <record> or <header> count, must be recordsExpectedPerRequest!
        if (recordCount != recordsExpectedPerRequest) {
          String message =
              OaiPmhTestUtilities.ERROR + ": " + recordOrHeader + " count mismatch, must be "
                  + recordsExpectedPerRequest + " if token is provided, but is " + recordCount;
          assertTrue(message, false);
        }
        if (size <= recordsExpectedPerRequest) {
          String message =
              OaiPmhTestUtilities.ERROR + ": completeListSize count mismatch, must be > "
                  + recordsExpectedPerRequest + " if token is provided, but is " + size;
          assertTrue(message, false);
        }
        if (cursor != recordsExpectedPerRequest * loopCount) {
          String message = OaiPmhTestUtilities.ERROR + ": cursor must be "
              + (loopCount * recordsExpectedPerRequest) + " in loop " + loopCount + ", but is "
              + cursor;
          assertTrue(message, false);
        }
      }
    }

    // If no token is provided, stop querying.
    else {
      // Check <record> count, must be completeListSize % recordsExpectedPerRequest.
      if (recordCount != size % recordsExpectedPerRequest) {
        System.err.println(theThreadName + OaiPmhTestUtilities.ERROR + ": " + recordOrHeader
            + " count mismatch, should be " + size % recordsExpectedPerRequest + ", but is "
            + recordCount);
        assertTrue(false);
      }

      // No resumption token available in response.
      return OaiPmhTestUtilities.NO_TOKEN;
    }

    return restok;
  }

  /**
   * @param theThreadName
   * @return
   */
  private static String tn(String theThreadName) {

    String result = theThreadName;

    if (theThreadName != null && !theThreadName.equals("") && !theThreadName.endsWith(" ")) {
      result = "[" + theThreadName + "] ";
    }

    return result;
  }

  /**
   * @param theResponseString
   * @param theMetadataFormat
   */
  private static void examineContent(String theResponseString, String theMetadataFormat) {

    // Check for correct metadata content according to metadata prefix.
    if (theMetadataFormat.equals(OAIDC_SCHEMA_FORMAT)
        && !theResponseString.contains("metadataPrefix=\"" + OAIDC_SCHEMA_FORMAT + "\"")
        && !theResponseString.contains(EXPECTED_OAIDC_FORMAT_CONTENT)) {
      System.out
          .println(OAIDC_SCHEMA_FORMAT + " needs to deliver content with schema: "
              + EXPECTED_OAIDC_FORMAT_CONTENT + "!");
      assertTrue(false);
    } else if (theMetadataFormat.equals(IDIOMMETS_SCHEMA_FORMAT)
        && !theResponseString.contains("metadataPrefix=\"" + IDIOMMETS_SCHEMA_FORMAT + "\"")
        && !theResponseString.contains(EXPECTED_IDIOMMETS_FORMAT_CONTENT)) {
      System.out
          .println(IDIOMMETS_SCHEMA_FORMAT + " needs to deliver content with schema: "
              + EXPECTED_IDIOMMETS_FORMAT_CONTENT + "!");
      assertTrue(false);
    } else if (theMetadataFormat.equals(DATACITE_SCHEMA_FORMAT)
        && !theResponseString.contains("metadataPrefix=\"" + DATACITE_SCHEMA_FORMAT + "\"")
        && !theResponseString.contains(EXPECTED_DATACITE_FORMAT_CONTENT)) {
      System.out
          .println(DATACITE_SCHEMA_FORMAT + " needs to deliver content with schema: "
              + EXPECTED_DATACITE_FORMAT_CONTENT + "!");
      assertTrue(false);
    }
  }

}
